/*
 *
 *
 */
#include "common.h"
#include "LPC214x.h"
#include "irq.h"
#include "usbreg.h"
#include "usbhw.h"

/*
 * functions
 */
void uhw_Init(void);
void uhw_Connect(int con);
void uhw_Reset(void);
void uhw_Suspend(void);
void uhw_Resume(void);
void uhw_WakeUp(void);
void uhw_SetAddress(u_char adr);
void uhw_Configure(int cfg);
void uhw_ConfigEP(u_char *pEPD);
void uhw_EnableEP(u_char EPNum);
void uhw_DisableEP(u_char EPNum);
void uhw_ResetEP(u_char EPNum);
void uhw_SetStallEP(u_char EPNum);
void uhw_ClrStallEP(u_char EPNum);
u_long uhw_WriteEP(u_char EPNum, u_char *pData, u_long cnt);
u_long uhw_ReadEP(u_char EPNum, u_char *pData);
static void ucmd_write_data(u_long cmd, u_long val);
static u_long ucmd_read_data(u_long cmd);
static void ucmd_no_data(u_long cmd);
static u_long get_phy_EPaddr(u_long EPNum);
static void USB_ISR(void) __attribute__ ((interrupt("IRQ")));

/*
 * externs
 */
/* in usbcore.c */
extern void ucore_isr_Ep0(u_long event);
/* in vcomuser.c */
extern const INTR_subFUNCs USB_Jump_Table[N_JT];
extern const INTR_subFUNCs USB_Jump_Table_EP[LPC_N_EP];


/* =============================================================== */

/*
 *  USB Initialize Function
 *    Return Value:    None
 */
void uhw_Init(void)
{
	/*
	 * Turn on USB
	 */
	PCONP |= 0x80000000;                      /* Turn On USB PCLK */

	/*
	 * Enable Vbus sense (P0.23)
	 */
	PINSEL1 |= (1<<14) | 0;

	/*
	 * select P0.31 pin function
	 *   1 - GoodLinc
	 *   2 - ~Connect
	 */
	PINSEL1 |= (2<<30) | 0;

	/* Configure 48MHz USB Clock;  FOsc = 12MHz, M = 4, P = 2 */
	PLL48CFG  = 0x23;                         /* M = 4, P = 2 */
	PLL48CON  = PLLCON_PLLE;                  /* PLL Enable */
	PLL48FEED = 0xAA;                         /* Feed Sequence 1 */
	PLL48FEED = 0x55;                         /* Feed Sequence 2 */

	while((PLL48STAT & PLLSTAT_PLOCK) == 0)   /* Wait for PLL Lock */
		;

	PLL48CON  = PLLCON_PLLE | PLLCON_PLLC;    /* PLL Enable & Connect */
	PLL48FEED = 0xAA;                         /* Feed Sequence 1 */
	PLL48FEED = 0x55;                         /* Feed Sequence 2 */

//	VICVectAddr0 = (unsigned long)USB_ISR;    /* USB Interrupt -> Vector 0 */
//	VICVectCntl0 = 0x20 | 22;                 /* USB Interrupt -> IRQ Slot 0 */
//	VICIntEnable = 1 << 22;                   /* Enable USB Interrupt */
	if(install_irq(USB_INT, (void *)USB_ISR) == FALSE) {
		return; //return(FALSE);
	}

	DEV_INT_EN = DEV_STAT_INT;            /* Enable Device Status Interrupt */

#if 1 /* Partial Manual Reset */
	uhw_Reset();
	uhw_SetAddress(0);
#endif

	/*
	 * Enable USB inerrupt
	 */
	USB_INT_STAT |= (1<<31);
}

/*
 *  USB Connect Function
 *    Parameters:      con:   Connect/Disconnect
 *    Return Value:    None
 */
void uhw_Connect(int con)
{
	ucmd_write_data(CMD_SET_DEV_STAT, DAT_WR_BYTE(con ? DEV_CON : 0));
}

/*
 *  USB Reset Function
 *    Return Value:    None
 */
void uhw_Reset(void)
{
rprintf("uRes\r\n");
	EP_INDEX = 0;
	MAXPACKET_SIZE = MY_MAX_PACKET_SIZE;

	EP_INDEX = 1;
	MAXPACKET_SIZE = MY_MAX_PACKET_SIZE;

	while((DEV_INT_STAT & EP_RLZED_INT) == 0)
		;

	EP_INT_CLR  = 0xFFFFFFFF;
	EP_INT_EN   = 0xFFFFFFFF;
	DEV_INT_CLR = 0xFFFFFFFF;
	DEV_INT_EN  = DEV_STAT_INT |
					((USB_Jump_Table[USB_JT_SOF] != NULL) ? FRAME_INT : 0) |
					((USB_Jump_Table[USB_JT_ERROR] != NULL) ? ERR_INT : 0) |
					EP_SLOW_INT;
}

/*
 *  USB Suspend Function
 *    Return Value:    None
 */
void uhw_Suspend(void)
{
rprintf("uSus\r\n");
  // Performed by Hardware
}

/*
 *  USB Resume Function
 *    Return Value:    None
 */
void uhw_Resume(void)
{
rprintf("uRsm\r\n");
  // Performed by Hardware
}

/*
 *  USB Remote Wakeup Function
 *    Return Value:    None
 */
void uhw_WakeUp(void)
{
	ucmd_write_data(CMD_SET_DEV_STAT, DAT_WR_BYTE(DEV_CON));
}

/*
 *  USB Set Address Function
 *    Parameters:      adr:   USB Address
 *    Return Value:    None
 */
void uhw_SetAddress(u_char adr)
{
	ucmd_write_data(CMD_SET_ADDR, DAT_WR_BYTE(DEV_EN | adr)); /* Don't wait for next */
	ucmd_write_data(CMD_SET_ADDR, DAT_WR_BYTE(DEV_EN | adr)); /*  Setup Status Phase */
}

/*
 *  USB Configure Function
 *    Parameters:      cfg:   Configure/Deconfigure
 *    Return Value:    None
 */
void uhw_Configure(int cfg)
{
	ucmd_write_data(CMD_CFG_DEV, DAT_WR_BYTE(cfg ? CONF_DVICE : 0));

	REALIZE_EP = 0x00000003;
	while((DEV_INT_STAT & EP_RLZED_INT) == 0)
		;
	DEV_INT_CLR = EP_RLZED_INT;
}

/*
 *  Configure USB Endpoint
 *    Parameters:      pEPD:  Pointer to Endpoint Descriptor
 *    Return Value:    None
 */
void uhw_ConfigEP(u_char *pEPD)
{
	u_long num;

	num = get_phy_EPaddr(pEPD[/*bEndpointAddress*/ 2]);
	REALIZE_EP |= (1 << num);
	EP_INDEX = num;
	MAXPACKET_SIZE = (pEPD[4]<<0) + (pEPD[5]<<8);	/* wMaxPacketSize */
	while((DEV_INT_STAT & EP_RLZED_INT) == 0)
		;
	DEV_INT_CLR = EP_RLZED_INT;
}

/*
 *  Enable USB Endpoint
 *    Parameters:      EPNum: Endpoint Number
 *                       EPNum.0..3: Address
 *                       EPNum.7:    Dir
 *    Return Value:    None
 */
void uhw_EnableEP(u_char EPNum)
{
	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(0));
}

/*
 *  Disable USB Endpoint
 *    Parameters:      EPNum: Endpoint Number
 *                       EPNum.0..3: Address
 *                       EPNum.7:    Dir
 *    Return Value:    None
 */
void uhw_DisableEP(u_char EPNum)
{
	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(EP_STAT_DA));
}

/*
 *  Reset USB Endpoint
 *    Parameters:      EPNum: Endpoint Number
 *                       EPNum.0..3: Address
 *                       EPNum.7:    Dir
 *    Return Value:    None
 */
void uhw_ResetEP(u_char EPNum)
{
	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(0));
}

/*
 *  Set Stall for USB Endpoint
 *    Parameters:      EPNum: Endpoint Number
 *                       EPNum.0..3: Address
 *                       EPNum.7:    Dir
 *    Return Value:    None
 */
void uhw_SetStallEP(u_char EPNum)
{
	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(EP_STAT_ST));
}

/*
 *  Clear Stall for USB Endpoint
 *    Parameters:      EPNum: Endpoint Number
 *                       EPNum.0..3: Address
 *                       EPNum.7:    Dir
 *    Return Value:    None
 */
void uhw_ClrStallEP(u_char EPNum)
{
	ucmd_write_data(CMD_SET_EP_STAT(get_phy_EPaddr(EPNum)), DAT_WR_BYTE(0));
}

/*
 *  Write USB Endpoint Data
 *    Parameters:      EPNum: Endpoint Number
 *                       EPNum.0..3: Address
 *                       EPNum.7:    Dir
 *                     pData: Pointer to Data Buffer
 *                     cnt:   Number of bytes to write
 *    Return Value:    Number of bytes written
 */
u_long uhw_WriteEP(u_char EPNum, u_char *pData, u_long cnt)
{
	u_long n, data;

	USB_CTRL = ((EPNum & 0x0F) << 2) | CTRL_WR_EN;

	TX_PLENGTH = cnt;

	for(n=0; n<(cnt+3)/4; n++) {
		data = (pData[0]<<0) + (pData[1]<<8) + (pData[2]<<16) + (pData[3]<<24);
		TX_DATA = data;
		pData += 4;
	}

	USB_CTRL = 0;

	ucmd_no_data(CMD_SEL_EP(get_phy_EPaddr(EPNum)));
	ucmd_no_data(CMD_VALID_BUF);

	return(cnt);
}

/*
 *  Read USB Endpoint Data
 *    Parameters:      EPNum: Endpoint Number
 *                       EPNum.0..3: Address
 *                       EPNum.7:    Dir
 *                     pData: Pointer to Data Buffer
 *    Return Value:    Number of bytes read
 */
u_long uhw_ReadEP(u_char EPNum, u_char *pData)
{
	u_long cnt, n, rx;

	USB_CTRL = ((EPNum & 0x0F) << 2) | CTRL_RD_EN;

	do {
		cnt = RX_PLENGTH;
	} while((cnt & PKT_RDY) == 0);
	cnt &= PKT_LNGTH_MASK;

	for(n=0; n<cnt /*&& i<DataSize*/; n++) {
		if((n % 4) == 0) {
			rx = RX_DATA;
		}
		*pData++ = rx & 0xff;
		rx >>= 8;
	}

	USB_CTRL = 0;

	ucmd_no_data(CMD_SEL_EP(get_phy_EPaddr(EPNum)));
	ucmd_no_data(CMD_CLR_BUF);

	return(cnt);
}

/*
 *  Write Command Data
 *    Parameters:      cmd:   Command
 *                     val:   Data
 *    Return Value:    None
 */
static void ucmd_write_data(u_long cmd, u_long val)
{
	CMD_CODE = cmd;
	while((DEV_INT_STAT & CCEMTY_INT) == 0)
		;
	DEV_INT_CLR = CCEMTY_INT;

	CMD_CODE = val;
	while((DEV_INT_STAT & CCEMTY_INT) == 0)
		;
	DEV_INT_CLR = CCEMTY_INT;
}

/*
 *  Read Command Data
 *    Parameters:      cmd:   Command
 *    Return Value:    Data Value
 */
static u_long ucmd_read_data(u_long cmd)
{
	u_long val;

	DEV_INT_CLR = CDFULL_INT;

	CMD_CODE = cmd;
	while((DEV_INT_STAT & CCEMTY_INT) == 0)
		;
	DEV_INT_CLR = CCEMTY_INT;

	while((DEV_INT_STAT & CDFULL_INT) == 0)
		;
	val = CMD_DATA;
	DEV_INT_CLR = CDFULL_INT;

	return(val);
}

/*
 *  Write Command
 *    Parameters:      cmd:   Command
 *    Return Value:    None
 */
static void ucmd_no_data(u_long cmd)
{
	CMD_CODE = cmd;
	while((DEV_INT_STAT & CCEMTY_INT) == 0)
		;
	DEV_INT_CLR = CCEMTY_INT;
}

/*
 *  Get Endpoint Address
 *    Parameters:      EPNum: Endpoint Number
 *                       EPNum.0..3: Address
 *                       EPNum.7:    Dir
 *    Return Value:    Endpoint Physical Address
 */
static u_long get_phy_EPaddr(u_long EPNum)
{
	u_long val;

	val = (EPNum & 0x0F) << 1;
	if(EPNum & 0x80) {
		val += 1;
	}
	return(val);
}

/* =============================================================== */

/*
 *  USB Interrupt Service Routine
 */
static void USB_ISR(void)
{
	u_long disr, eisr, val, n;

	disr = DEV_INT_STAT;                      /* Device Interrupt Status */
rprintf("DEV_INT_STAT:%08x\r\n", disr);
	if(disr & DEV_STAT_INT) {
cm_putchar('D');
		/*
		 * Device Status Interrupt (Reset, Suspend/Resume, Connect change)
		 *   get device status detail
		 */
		ucmd_no_data(CMD_GET_DEV_STAT);
		val = ucmd_read_data(DAT_GET_DEV_STAT);
rprintf("val:%08x\r\n", val);

		if(val & DEV_RST) {                    /* Reset */
			uhw_Reset();
			if(USB_Jump_Table[USB_JT_RESET] != NULL) {
				USB_Jump_Table[USB_JT_RESET](0);
			}
			goto isr_end;
		}

		if(val & DEV_SUS_CH) {                 /* Suspend/Resume */
			if(val & DEV_SUS) {                  /* Suspend */
				uhw_Suspend();
				if(USB_Jump_Table[USB_JT_SUSPEND] != NULL) {
					USB_Jump_Table[USB_JT_SUSPEND](0);
				}
			} else {                              /* Resume */
				uhw_Resume();
				if(USB_Jump_Table[USB_JT_RESUME] != NULL) {
					USB_Jump_Table[USB_JT_RESUME](0);
				}
			}

			goto isr_end;
		}

		if(val & DEV_CON_CH) {                 /* Connect change */
			if(USB_Jump_Table[USB_JT_POWER] != NULL) {
				USB_Jump_Table[USB_JT_POWER](val & DEV_CON);
			}
			goto isr_end;
		}
	}

	if(disr & FRAME_INT) {
		/*
		 * Start of Frame Interrupt
		 */
		if(USB_Jump_Table[USB_JT_SOF] != NULL) {
			ucmd_no_data(CMD_RD_FRAME);
			val = ucmd_read_data(DAT_RD_FRAME);
			val = val | (ucmd_read_data(DAT_RD_FRAME) << 8);
			USB_Jump_Table[USB_JT_SOF](val);
		}
	}

	if(disr & ERR_INT) {
		/*
		 * Error Interrupt
		 */
		if(USB_Jump_Table[USB_JT_ERROR] != NULL) {
			ucmd_no_data(CMD_RD_ERR_STAT);
			val = ucmd_read_data(DAT_RD_ERR_STAT);

			if        (val & 0x01) {
				val = USB_ERR_PID;
			} else if(val & 0x02) {
				val = 0x100;	// Unexpected Error
			} else if(val & 0x04) {
				val = USB_ERR_CRC;
			} else if(val & 0x08) {
				val = USB_ERR_TIMEOUT;
			} else if(val & 0x10) {
				val = USB_ERR_EOP;
			} else if(val & 0x20) {
				val = 0x101;	// Buffer Overrun
			} else if(val & 0x40) {
				val = USB_ERR_BIT_STUFF;
			} else if(val & 0x80) {
				val = USB_ERR_DATA_TOGGLE;
			}
			USB_Jump_Table[USB_JT_ERROR](val);
		}
	}

	if(disr & EP_SLOW_INT) {
		/*
		 * Ednpoint's Slow Interrupt
		 */
		while((eisr = EP_INT_STAT)) {           /* Endpoint Interrupt Status */
			/*
			 * Check All Endpoints
			 */
			for(n=0; n<LPC_N_EP; n++) {
				if(eisr & (1 << n)) {
					EP_INT_CLR = 1 << n;
					while((DEV_INT_STAT & CDFULL_INT) == 0)
						;
					val = CMD_DATA;
					DEV_INT_CLR = CDFULL_INT;

					if(n == 0) {
						/* control out */
						if(val & EP_SEL_STP) {         /* Setup Packet */
							ucore_isr_Ep0(USB_EVT_SETUP);
						} else {
							ucore_isr_Ep0(USB_EVT_OUT);
						}
					} else if(n == 1) {
						/* control in */
						ucore_isr_Ep0(USB_EVT_IN);
					} else {
						/* generic endpoints */
						if(USB_Jump_Table_EP[n] != NULL) {
							if((n & 1) == 0) {
								USB_Jump_Table_EP[n](USB_EVT_OUT);
							} else {
								USB_Jump_Table_EP[n](USB_EVT_IN);
							}
						}
					}
				}
			}
		}
	}

isr_end:
	DEV_INT_CLR = disr;

	VICVectAddr = 0;				/* Acknowledge Interrupt */
}
